/*
 * Copyright 2005-2007 Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*
 *
 * Trivial CRT ViaOutput implementation.
 *
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "via_driver.h"

#ifndef _XF86_ANSIC_H
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#endif

#include "via_regs.h"

#include "via_id.h"

#include "xf86i2c.h"
#include "xf86DDC.h"

#include "via_output.h"

/*
 * Output->Private
 */
struct CRTPrivate {
    Bool Sense; /* used for init only */

    int BandWidth; /* Limit modes - Config or EDID provided. */
};

/*
 *
 */
static void
CRTPrivateDestroy(struct ViaOutput *Output)
{
    free(Output->Private);

    /* So we won't try to destroy this twice */
    Output->PrivateDestroy = NULL;
}

/*
 *
 */
static struct CRTPrivate *
CRTPrivateCreate(struct ViaOutput *Output)
{
    VIAFUNC(Output);

    Output->PrivSize = sizeof(struct CRTPrivate);
    Output->Private =  xnfcalloc(1, Output->PrivSize);
    memset(Output->Private, 0, Output->PrivSize);

    Output->PrivateDestroy = CRTPrivateDestroy;

    return Output->Private;
}

/*
 * Option handling.
 */
enum CRTOpts {
    OPTION_CRTSENSE,
    OPTION_CRTBANDWIDTH,
    OPTION_CRTEDID
};

static OptionInfoRec CRTOptions[] =
{
    {OPTION_CRTSENSE,     "CRTSense",     OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_CRTBANDWIDTH, "CRTBandWidth", OPTV_INTEGER, {0}, FALSE},
    {OPTION_CRTEDID,      "CRTEDID",      OPTV_ANYSTR,  {0}, FALSE},
    {-1,  NULL,  OPTV_NONE,  {0},  FALSE }
};

/*
 *
 */
static OptionInfoPtr
CRTGetOptions(struct ViaOutput *Output)
{
    struct CRTPrivate *Private = (struct CRTPrivate *) Output->Private;
    OptionInfoPtr  Options;
    char *EDIDFileName;

    Options = xnfalloc(sizeof(CRTOptions));
    memcpy(Options, CRTOptions, sizeof(CRTOptions));

    xf86ProcessOptions(Output->scrnIndex,
		       xf86Screens[Output->scrnIndex]->options, Options);

    if (!xf86ReturnOptValBool(Options, OPTION_CRTSENSE, TRUE)) {
        xf86DrvMsg(Output->scrnIndex, X_WARNING, "Disabling CRT Sensing."
                   " CRT is considered always attached.\n");
        Private->Sense = FALSE;
    } else
        Private->Sense = TRUE;

    /* TODO, clean this up -- move to Output->MaxClock */
    Private->BandWidth = 0;
    xf86GetOptValInteger(Options, OPTION_CRTBANDWIDTH, &Private->BandWidth);
    if (Private->BandWidth && (Private->BandWidth < 1000)) {
        xf86DrvMsg(Output->scrnIndex, X_WARNING, "%s: Option CRTBandWidth needs "
                   "to be provided in kHz not MHz.\n", __func__);
        Private->BandWidth = 0;
    }


    EDIDFileName = xf86GetOptValString(Options, OPTION_CRTEDID);
    if (EDIDFileName) {
        FILE *EDIDFile;

        EDIDFile = fopen(EDIDFileName, "r");
        if (!EDIDFile)
            xf86DrvMsg(Output->scrnIndex, X_WARNING,
                       "%s: Unable to open EDID file \"%s\": %s\n",
                       __func__, EDIDFileName, strerror(errno));
        else {
            unsigned char Block[128];

            if (fread(Block, 1, 128, EDIDFile) != 128)
                 xf86DrvMsg(Output->scrnIndex, X_WARNING,
                       "%s: Unable to read from EDID file \"%s\": %s\n",
                            __func__, EDIDFileName, strerror(errno));
            else
                Output->EDID = xf86InterpretEDID(Output->scrnIndex, Block);

	    fclose(EDIDFile);
        }
    }

    free(Options);
    return NULL;
}

/*
 *
 * Massive amount of code to detect whether a CRT is present.
 *
 * This is done the old style, by creating some overscan, setting up the
 * overscan colour, then waiting for VSync/Blanking to end, and then
 * sampling the Switch Sense bit.
 */
/*
 * Check whether the VSync and Blanking bits are changing.
 */
static Bool
VGACRTSenseSignals(struct ViaOutput *Output)
{
    Bool VSync, VSyncChanged = 0;
    Bool Blank, BlankChanged = 0;
    CARD8 tmp;
    int i;

    tmp = VGAStatus1Read(Output);
    VSync = tmp & 0x08;
    Blank = tmp & 0x01;

    for (i = 0; i < 100000; i++) {
	tmp = VGAStatus1Read(Output);

	if ((tmp & 0x08) != VSync)
	    VSyncChanged = 1;

	if ((tmp & 0x01) != Blank)
	    BlankChanged = 1;

	if (VSyncChanged && BlankChanged)
	    return TRUE;
    }

    return FALSE;
}

/*
 * This code checks for the sense bit in several passes. This gives the
 * monitor to be tested some time to sync up and start drawing some current,
 * but also doesn't waste too much time.
 */
static Bool
VGACRTSenseLowlevel(struct ViaOutput *Output)
{
    int i;

    for (i = 0; i < 0x10; i++) {
	/* Wait for VSync to start */
	while (!(VGAStatus1Read(Output) & 0x08))
	    ;

	/* Wait for blanking to end */
	while ((VGAStatus1Read(Output) & 0x01))
	    ;

	/* slight delay: wait for the opamps to catch up. */
	inb(VGAStatus0Read(Output));

	if (VGAStatus0Read(Output) & 0x10) {
	    ViaDebug(Output->scrnIndex,
		     "This detection required %d passes.\n", i);
	    return TRUE;
	}
    }

    return FALSE;
}

/*
 * Just alter the running mode slightly.
 */
static Bool
VT3122AxSenseEasy(struct ViaOutput *Output)
{
    CARD8 SR16, SR1A, CR16, CR36, AR11, DAC00[3];
    CARD8 tmp;
    Bool Found;

    AR11 = VGAARRead(Output, 0x11);
    SR16 = VGASRRead(Output, 0x16);
    SR1A = VGASRRead(Output, 0x1A);
    CR16 = VGACRRead(Output, 0x16);
    CR36 = VGACRRead(Output, 0x36);

    /* Make sure that we set the right overscan colour */
    VGAARWrite(Output, 0x11, 0x00);
    VGADACAccessDisable(Output);

    VGASRMask(Output, 0x1A, 0x00, 0x01); /* Upload DAC to CRTC1 */

    VGACRMask(Output, 0x36, 0x00, 0x30); /* Sync */

    tmp = VGACRRead(Output, 0x10) & 0xF0; /* Get low byte of VSyncStart */
    tmp |= VGACRRead(Output, 0x11) & 0x0F; /* Get VSyncEnd */
    if (tmp < VGACRRead(Output, 0x10))
	tmp += 0x10;
    VGACRWrite(Output, 0x16, tmp - 1);

    if (VGASRRead(Output, 0x15) & 0x0C) /* Are we using > 8 bpp? */
	VGASRMask(Output, 0x16, 0x80, 0x80); /* enable gamma */

    /* store original dac 0x00 value */
    VGADACReadAddressWrite(Output, 0x00);
    DAC00[0] = VGADACDataRead(Output);
    DAC00[1] = VGADACDataRead(Output);
    DAC00[2] = VGADACDataRead(Output);

    /* Now write our new overscan colour */
    VGADACWriteAddressWrite(Output, 0x00);
    VGADACDataWrite(Output, 0x55);
    VGADACDataWrite(Output, 0x55);
    VGADACDataWrite(Output, 0x55);

    Found = VGACRTSenseLowlevel(Output);

    VGADACWriteAddressWrite(Output, 0x00);
    VGADACDataWrite(Output, DAC00[0]);
    VGADACDataWrite(Output, DAC00[1]);
    VGADACDataWrite(Output, DAC00[2]);

    VGAARWrite(Output, 0x11, AR11);
    VGADACAccessDisable(Output);

    VGASRWrite(Output, 0x16, SR16);
    VGASRWrite(Output, 0x1A, SR1A);
    VGACRWrite(Output, 0x16, CR16);
    VGACRWrite(Output, 0x36, CR36);

    return Found;
}

/*
 * We need to set up a full mode here, from absolute scratch too.
 */
static Bool
VT3122AxSensePainful(struct ViaOutput *Output)
{
    CARD8 Enable, Misc;
    CARD8 SR[0x06], SR10, SR15, SR16, SR17, SR18, SR1A;
    CARD8 CR[0x19], CR33, CR35, CR36, CR6A, CR6B, CR6C;
    CARD8 AR[0x15], GR[0x09], DAC[3];
    Bool Found;
    int i;

    /* Some working mode tables. */
    CARD8 SRTable[0x05] = { 0x03, 0x00, 0x03, 0x00, 0x02 };
    CARD8 CRTable[0x19] = { 0x5F, 0x4F, 0x50, 0x82, 0x55, 0x81, 0xBF, 0x1F,
			    0x00, 0x4F, 0x0E, 0x0E, 0x00, 0x00, 0x00, 0x00,
			    0x9C, 0x0E, 0x8F, 0x28, 0x00, 0x96, 0x9D, 0xA3, 0xFF };
    CARD8 GRTable[0x09] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x0E, 0x00, 0xFF };
    CARD8 ARTable[0x15] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
			    0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
			    0x0C, 0x00, 0x0F, 0x08, 0x00 };

    Enable = VGAEnableRead(Output);
    VGAEnableMask(Output, 0x01, 0x01);

    Misc = VGAMiscRead(Output);
    VGAMiscMask(Output, 0x23, 0x23);

    for (i = 0; i < 6; i++)
	SR[i] = VGASRRead(Output, i);
    SR10 = VGASRRead(Output, 0x10);
    SR15 = VGASRRead(Output, 0x15);
    SR16 = VGASRRead(Output, 0x16);
    SR17 = VGASRRead(Output, 0x17);
    SR18 = VGASRRead(Output, 0x18);
    SR1A = VGASRRead(Output, 0x1A);

    VGASRMask(Output, 0x10, 0x01, 0x01);
    VGASRMask(Output, 0x1A, 0x40, 0x40);

    for (i = 0; i < 0x19; i++)
	CR[i] = VGACRRead(Output, i);
    VGACRMask(Output, 0x11, 0x00, 0x80);

    CR33 = VGACRRead(Output, 0x33);
    CR35 = VGACRRead(Output, 0x35);
    CR36 = VGACRRead(Output, 0x36);
    CR6A = VGACRRead(Output, 0x6A);
    CR6B = VGACRRead(Output, 0x6B);
    CR6C = VGACRRead(Output, 0x6C);

    for (i = 0; i < 0x09; i++)
	GR[i] = VGAGRRead(Output, i);

    for (i = 0; i < 0x15; i++)
	AR[i] = VGAARRead(Output, i);

    VGADACReadAddressWrite(Output, 0x00);
    DAC[0] = VGADACDataRead(Output);
    DAC[1] = VGADACDataRead(Output);
    DAC[2] = VGADACDataRead(Output);

    /* Setup our new mode */
    for (i = 0; i < 0x05; i++)
	VGASRWrite(Output, i, SRTable[i]);
    VGASRMask(Output, 0x15, 0x00, 0x22);
    VGASRWrite(Output, 0x16, 0x08);
    VGASRWrite(Output, 0x17, 0x1F);
    VGASRWrite(Output, 0x18, 0x4E);
    VGASRMask(Output, 0x1A, 0x00, 0x09);

    for (i = 0; i < 0x19; i++)
	VGACRWrite(Output, i, CRTable[i]);
    VGACRWrite(Output, 0x33, 0x00);
    VGACRWrite(Output, 0x35, 0x00);
    VGACRMask(Output, 0x36, 0x00, 0x30);

    VGAMiscMask(Output, 0x44, 0xCC);

    for (i = 0; i < 0x09; i++)
	VGAGRWrite(Output, i, GRTable[i]);

    for (i = 0; i < 0x15; i++)
	VGAARWrite(Output, i, ARTable[i]);
    VGADACAccessDisable(Output);

    VGADACMaskWrite(Output, 0xFF);

    VGASRMask(Output, 0x00, 0x02, 0x02);

    /* Now start detection */

    /* Now write our new overscan colour */
    VGADACWriteAddressWrite(Output, 0x00);
    VGADACDataWrite(Output, 0x55);
    VGADACDataWrite(Output, 0x55);
    VGADACDataWrite(Output, 0x55);

    Found = VGACRTSenseLowlevel(Output);

    /* Restore */
    VGADACWriteAddressWrite(Output, 0x00);
    VGADACDataWrite(Output, DAC[0]);
    VGADACDataWrite(Output, DAC[1]);
    VGADACDataWrite(Output, DAC[2]);

    for (i = 0; i < 0x15; i++)
	VGAARWrite(Output, i, AR[i]);
    VGADACAccessDisable(Output);

    for (i = 0; i < 0x09; i++)
	VGAGRWrite(Output, i, GR[i]);

    for (i = 0; i < 0x19; i++)
	VGACRWrite(Output, i, CR[i]);
    VGACRWrite(Output, 0x33, CR33);
    VGACRWrite(Output, 0x35, CR35);
    VGACRWrite(Output, 0x36, CR36);
    VGACRWrite(Output, 0x6A, CR6A);
    VGACRWrite(Output, 0x6B, CR6B);
    VGACRWrite(Output, 0x6C, CR6C);

    for (i = 0; i < 6; i++)
	VGASRWrite(Output, i, SR[i]);
    VGASRWrite(Output, 0x10, SR10);
    VGASRWrite(Output, 0x15, SR15);
    VGASRWrite(Output, 0x16, SR16);
    VGASRWrite(Output, 0x17, SR17);
    VGASRWrite(Output, 0x18, SR18);
    VGASRWrite(Output, 0x1A, SR1A);

    VGAMiscWrite(Output, Misc);
    VGAEnableWrite(Output, Enable);

    return Found;
}

/*
 * This is the painful one, does CRT sensing the old VGA way.
 */
static Bool
VT3122AxCRTSense(struct ViaOutput *Output)
{
    Bool Found;

    VIAFUNC(Output);

    if (!VGACRTSenseSignals(Output) || (VGACRRead(Output, 0x6A) & 0x80) ||
	(VGASRRead(Output, 0x16) & 0x40))
	Found = VT3122AxSensePainful(Output);
    else
	Found = VT3122AxSenseEasy(Output);

    if (!Found)
	xf86DrvMsg(Output->scrnIndex, X_PROBED, "%s: Nothing connected.\n",
		   Output->Name);
    else
	xf86DrvMsg(Output->scrnIndex, X_PROBED, "%s: Attached!\n",
		   Output->Name);

    return Found;
}

/*
 * CR40.7 is not functional on the VT3122Ax, so use the other function there.
 */
static Bool
VT3122CxCRTSense(struct ViaOutput *Output)
{
    CARD8 SR01, CR36;
    Bool Found = FALSE;

    VIAFUNC(Output);

    SR01 = VGASRRead(Output, 0x01);
    CR36 = VGACRRead(Output, 0x36);

    VGASRMask(Output, 0x01, 0x00, 0x20);
    VGACRMask(Output, 0x36, 0x00, 0xF0);

    VGASRMask(Output, 0x40, 0x80, 0x80);

    usleep(1);

    if (VGAStatus0Read(Output) & 0x10)
        Found = TRUE;

    VGASRMask(Output, 0x40, 0x00, 0x80);

    VGASRWrite(Output, 0x01, SR01);
    VGACRWrite(Output, 0x36, CR36);

    if (!Found)
	xf86DrvMsg(Output->scrnIndex, X_PROBED, "%s: Nothing connected.\n",
		   Output->Name);
    else
	xf86DrvMsg(Output->scrnIndex, X_PROBED, "%s: Attached!\n",
		   Output->Name);

    return Found;
}

/*
 * Sync/Unsync
 */
static void
CRTPower(struct ViaOutput *Output, Bool On)
{
    VIAFUNC(Output);

    if (On)
        VGACRMask(Output, 0x36, 0x00, 0x30);
    else
        VGACRMask(Output, 0x36, 0x30, 0x30);
}

/*
 * Empty.
 */
static ModeStatus
CRTModeValid(struct ViaOutput *Output, DisplayModePtr mode)
{
    return MODE_OK;
}

/*
 * Empty.
 */
static void
CRTMode(struct ViaOutput *Output, DisplayModePtr mode)
{
    ;
}

/*
 *
 */
struct ViaOutput *
ViaCRTInit(int scrnIndex, I2CDevPtr pDev)
{
    VIAPtr pVia = VIAPTR(xf86Screens[scrnIndex]);
    struct ViaOutput *Output;
    struct CRTPrivate *Private;

    VIAFUNC(pVia);

    if (pDev)
        xf86DrvMsg(scrnIndex, X_WARNING,
                   "%s: Ignoring I2C Device.\n", __func__);

    Output = xnfcalloc(1, sizeof(struct ViaOutput));
    memset(Output, 0, sizeof(struct ViaOutput));

    Output->scrnIndex = scrnIndex;
    Output->Type = OUTPUT_CRT;
    Output->Name = "CRT";

    Output->Active = TRUE;
    Output->ClockMaster = FALSE;

    Private = CRTPrivateCreate(Output);

    Output->Options = CRTGetOptions(Output);

    if (Private->Sense) {
	if ((pVia->Chipset == VT3122) && VT3122_REV_IS_AX(pVia->HostRev))
	    Output->Sense = VT3122AxCRTSense;
	else
	    Output->Sense = VT3122CxCRTSense;
    }

    Output->ModeValid = CRTModeValid;
    Output->Mode = CRTMode;
    Output->Power = CRTPower;

    return Output;
}
