/*
 * Copyright 2004-2009 Luc Verhaegen.
 * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
/*
 * Implements five i2c busses through registers SR25, SR26, SR2c, SR31 and SR3D
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "xf86i2c.h"

#include "via_driver.h"
#include "via_regs.h"
#include "via_id.h"

#define SDA_CONTROL 0x40
#define SDA_WRITE 0x10
#define SDA_READ  0x04

#define SCL_CONTROL 0x80
#define SCL_WRITE 0x20
#define SCL_READ  0x08

/*
 *
 * I2C Bus 1 at SR26.
 *
 */
/*
 *
 */
static void
I2CBus1PutBits(I2CBusPtr Bus, int clock,  int data)
{
    CARD8 value = 0x01; /* Enable */

    if (clock)
        value |= SCL_WRITE;

    if (data)
        value |= SDA_WRITE;

    VGASRMask(Bus, 0x26, value, 0x01 | SCL_WRITE | SDA_WRITE);
}

/*
 *
 */
static void
I2CBus1GetBits(I2CBusPtr Bus, int *clock, int *data)
{
    CARD8 value = VGASRRead(Bus, 0x26);

    *clock = (value & SCL_READ) != 0;
    *data  = (value & SDA_READ) != 0;
}

/*
 *
 */
static I2CBusPtr
I2CBus1Init(int scrnIndex)
{
    I2CBusPtr Bus = xf86CreateI2CBusRec();

    if (!Bus)
	return NULL;

    Bus->BusName    = "Bus 1";
    Bus->scrnIndex  = scrnIndex;
    Bus->I2CPutBits = I2CBus1PutBits;
    Bus->I2CGetBits = I2CBus1GetBits;

    if (!xf86I2CBusInit(Bus)) {
        xf86DestroyI2CBusRec(Bus, TRUE, FALSE);
        return NULL;
    }

    return Bus;
}

/*
 *
 * I2C Bus 2 at SR31.
 *
 */
/*
 *
 */
static void
I2CBus2PutBits(I2CBusPtr Bus, int clock,  int data)
{
    CARD8 value = 0x01; /* Enable */

    if (clock)
        value |= SCL_WRITE;

    if (data)
        value |= SDA_WRITE;

    VGASRMask(Bus, 0x31, value, 0x01 | SCL_WRITE | SDA_WRITE);
}

/*
 *
 */
static void
I2CBus2GetBits(I2CBusPtr Bus, int *clock, int *data)
{
    CARD8 value = VGASRRead(Bus, 0x31);

    *clock = (value & SCL_READ) != 0;
    *data  = (value & SDA_READ) != 0;
}

/*
 *
 */
static I2CBusPtr
I2CBus2Init(int scrnIndex)
{
    I2CBusPtr Bus = xf86CreateI2CBusRec();

    if (!Bus)
	return NULL;

    Bus->BusName    = "Bus 2";
    Bus->scrnIndex  = scrnIndex;
    Bus->I2CPutBits = I2CBus2PutBits;
    Bus->I2CGetBits = I2CBus2GetBits;

    if (!xf86I2CBusInit(Bus)) {
        xf86DestroyI2CBusRec(Bus, TRUE, FALSE);
        return NULL;
    }

    return Bus;
}

/*
 *
 * I2C Bus 3 at SR25.
 *
 */
/*
 *
 */
static void
I2CBus3PutBits(I2CBusPtr Bus, int clock,  int data)
{
    CARD8 value = 0xC0; /* enable control lines for sda/scl */

    if (clock)
        value |= SCL_WRITE;

    if (data)
        value |= SDA_WRITE;

    VGASRMask(Bus, 0x25, value, 0xC0 | SCL_WRITE | SDA_WRITE);
}

/*
 *
 */
static void
I2CBus3GetBits(I2CBusPtr Bus, int *clock, int *data)
{
    CARD8 value = VGASRRead(Bus, 0x25);

    *clock = (value & SCL_READ) != 0;
    *data  = (value & SDA_READ) != 0;
}

/*
 *
 */
static I2CBusPtr
I2CBus3Init(int scrnIndex)
{
    I2CBusPtr Bus = xf86CreateI2CBusRec();

    if (!Bus)
	return NULL;

    Bus->BusName    = "Bus 3";
    Bus->scrnIndex  = scrnIndex;
    Bus->I2CPutBits = I2CBus3PutBits;
    Bus->I2CGetBits = I2CBus3GetBits;

    if (!xf86I2CBusInit(Bus)) {
        xf86DestroyI2CBusRec(Bus, TRUE, FALSE);
        return NULL;
    }

    return Bus;
}

/*
 *
 * I2C Bus 4 at SR2C.
 *
 */
/*
 *
 */
static void
I2CBus4PutBits(I2CBusPtr Bus, int clock,  int data)
{
    CARD8 value = 0xC0; /* enable control lines for sda/scl */

    if (clock)
        value |= SCL_WRITE;

    if (data)
        value |= SDA_WRITE;

    VGASRMask(Bus, 0x2C, value, 0xC0 | SCL_WRITE | SDA_WRITE);
}

/*
 *
 */
static void
I2CBus4GetBits(I2CBusPtr Bus, int *clock, int *data)
{
    CARD8 value = VGASRRead(Bus, 0x2C);

    *clock = (value & SCL_READ) != 0;
    *data  = (value & SDA_READ) != 0;
}

/*
 *
 */
static I2CBusPtr
I2CBus4Init(int scrnIndex)
{
    I2CBusPtr Bus = xf86CreateI2CBusRec();

    if (!Bus)
	return NULL;

    Bus->BusName    = "Bus 4";
    Bus->scrnIndex  = scrnIndex;
    Bus->I2CPutBits = I2CBus4PutBits;
    Bus->I2CGetBits = I2CBus4GetBits;

    if (!xf86I2CBusInit(Bus)) {
        xf86DestroyI2CBusRec(Bus, TRUE, FALSE);
        return NULL;
    }

    return Bus;
}

/*
 *
 * I2C Bus 5 at SR3D.
 *
 */
/*
 *
 */
static void
I2CBus5PutBits(I2CBusPtr Bus, int clock,  int data)
{
    CARD8 value = 0xC0; /* enable control lines for sda/scl */

    if (clock)
        value |= SCL_WRITE;

    if (data)
        value |= SDA_WRITE;

    VGASRMask(Bus, 0x3D, value, 0xC0 | SCL_WRITE | SDA_WRITE);
}

/*
 *
 */
static void
I2CBus5GetBits(I2CBusPtr Bus, int *clock, int *data)
{
    CARD8 value = VGASRRead(Bus, 0x3D);

    *clock = (value & SCL_READ) != 0;
    *data  = (value & SDA_READ) != 0;
}

/*
 *
 */
static I2CBusPtr
I2CBus5Init(int scrnIndex)
{
    I2CBusPtr Bus = xf86CreateI2CBusRec();

    if (!Bus)
	return NULL;

    Bus->BusName    = "Bus 5";
    Bus->scrnIndex  = scrnIndex;
    Bus->I2CPutBits = I2CBus5PutBits;
    Bus->I2CGetBits = I2CBus5GetBits;

    if (!xf86I2CBusInit(Bus)) {
        xf86DestroyI2CBusRec(Bus, TRUE, FALSE);
        return NULL;
    }

    return Bus;
}

/*
 * Separate bus infrastructure for the VT7205.
 * Uses the same gpio ports as bus4 on later chips.
 */
static void
VT7205I2CBus3SCLSet(I2CBusPtr Bus, Bool SCL)
{
    if (SCL)
	VGASRMask(Bus, 0x2C, SCL_CONTROL | SCL_WRITE, SCL_CONTROL | SCL_WRITE);
    else
	VGASRMask(Bus, 0x2C, SCL_CONTROL, SCL_CONTROL | SCL_WRITE);
}

static void
VT7205I2CBus3SDASet(I2CBusPtr Bus, Bool SDA)
{
    if (SDA)
	VGASRMask(Bus, 0x2C, SDA_CONTROL | SDA_WRITE, SDA_CONTROL | SDA_WRITE);
    else
	VGASRMask(Bus, 0x2C, SDA_CONTROL, SDA_CONTROL | SDA_WRITE);
}

static void
VT7205I2CBus3SDARelease(I2CBusPtr Bus)
{
    VGASRMask(Bus, 0x2C, 0, SDA_CONTROL);
}

static Bool
VT7205I2CBus3SDARead(I2CBusPtr Bus)
{
    if (VGASRRead(Bus, 0x2C) & SDA_READ)
	return TRUE;
    else
	return FALSE;
}

/*
 *
 */
static Bool
VT7205I2CBus3Start(I2CBusPtr Bus, int timeout)
{
    VT7205I2CBus3SDASet(Bus, TRUE);
    VT7205I2CBus3SCLSet(Bus, TRUE);
    Bus->I2CUDelay(Bus, Bus->RiseFallTime);

    VT7205I2CBus3SDASet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);

    VT7205I2CBus3SCLSet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);

    return TRUE;
}

/*
 *
 */
static void
VT7205I2CBus3Stop(I2CDevPtr Dev)
{
    I2CBusPtr Bus = Dev->pI2CBus;

    VT7205I2CBus3SDASet(Bus, FALSE);
    VT7205I2CBus3SCLSet(Bus, TRUE);
    Bus->I2CUDelay(Bus, Bus->RiseFallTime);

    VT7205I2CBus3SDASet(Bus, TRUE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);

    VT7205I2CBus3SCLSet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);
}

/*
 *
 */
static Bool
VT7205I2CBus3PutByte(I2CDevPtr Dev, I2CByte data)
{
    I2CBusPtr Bus = Dev->pI2CBus;
    Bool ret;
    int i;

    VT7205I2CBus3SCLSet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->RiseFallTime);

    for (i = 7; i >= 0; i--) {
	if ((data >> i) & 0x01)
	    VT7205I2CBus3SDASet(Bus, TRUE);
	else
	    VT7205I2CBus3SDASet(Bus, FALSE);
	Bus->I2CUDelay(Bus, Bus->RiseFallTime);

	VT7205I2CBus3SCLSet(Bus, TRUE);
	Bus->I2CUDelay(Bus, Bus->HoldTime);

	VT7205I2CBus3SCLSet(Bus, FALSE);
	Bus->I2CUDelay(Bus, Bus->HoldTime);
    }

    /* raise first to avoid false positives */
    VT7205I2CBus3SDASet(Bus, TRUE);
    VT7205I2CBus3SDARelease(Bus);
    Bus->I2CUDelay(Bus, Bus->RiseFallTime);

    VT7205I2CBus3SCLSet(Bus, TRUE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);

    ret = !VT7205I2CBus3SDARead(Bus);

    VT7205I2CBus3SCLSet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);

    return ret;
}

/*
 *
 */
static Bool
VT7205I2CBus3GetByte(I2CDevPtr Dev, I2CByte *data, Bool last)
{
    I2CBusPtr Bus = Dev->pI2CBus;
    int i;

    *data = 0;

    VT7205I2CBus3SCLSet(Bus, FALSE);
    VT7205I2CBus3SDARelease(Bus);
    Bus->I2CUDelay(Bus, Bus->RiseFallTime);

    for (i = 7; i >= 0; i--) {
	VT7205I2CBus3SCLSet(Bus, TRUE);
	Bus->I2CUDelay(Bus, Bus->HoldTime);

	if (VT7205I2CBus3SDARead(Bus))
	    *data |= 0x01 << i;
	Bus->I2CUDelay(Bus, Bus->RiseFallTime);

	VT7205I2CBus3SCLSet(Bus, FALSE);
	Bus->I2CUDelay(Bus, Bus->HoldTime);
    }

    if (last) /* send NACK */
	VT7205I2CBus3SDASet(Bus, TRUE);
    else /* send ACK */
	VT7205I2CBus3SDASet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->RiseFallTime);

    VT7205I2CBus3SCLSet(Bus, TRUE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);
    VT7205I2CBus3SCLSet(Bus, FALSE);
    Bus->I2CUDelay(Bus, Bus->HoldTime);

    return TRUE;
}

static Bool
VT7205I2CBus3Address(I2CDevPtr Dev, I2CSlaveAddr addr)
{
    I2CBusPtr Bus = Dev->pI2CBus;

    if (Bus->I2CStart(Dev->pI2CBus, Dev->StartTimeout)) {
	if (Bus->I2CPutByte(Dev, addr & 0xFF)) {
	    if ((addr & 0xF8) != 0xF0 &&
		(addr & 0xFE) != 0x00)
		return TRUE;

	    if (Bus->I2CPutByte(Dev, (addr >> 8) & 0xFF))
		return TRUE;
	}

	Bus->I2CStop(Dev);
    }

    return FALSE;
}

/*
 *
 */
static I2CBusPtr
VT7205I2CBus3Init(int scrnIndex)
{
    I2CBusPtr Bus = xf86CreateI2CBusRec();

    if (!Bus)
	return NULL;

    Bus->BusName    = "Bus 3";
    Bus->scrnIndex  = scrnIndex;

    Bus->I2CPutByte = VT7205I2CBus3PutByte;
    Bus->I2CGetByte = VT7205I2CBus3GetByte;
    Bus->I2CAddress = VT7205I2CBus3Address;
    Bus->I2CStop = VT7205I2CBus3Stop;
    Bus->I2CStart = VT7205I2CBus3Start;

    if (!xf86I2CBusInit(Bus)) {
        xf86DestroyI2CBusRec(Bus, TRUE, FALSE);
        return NULL;
    }

    return Bus;
}

/*
 *
 */
void
ViaI2CInit(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);

    VIAFUNC(pScrn);

    pVia->I2CBus[0] = I2CBus1Init(pScrn->scrnIndex);
    pVia->I2CBus[1] = I2CBus2Init(pScrn->scrnIndex);

    if (pVia->Chipset == VT7205)
	pVia->I2CBus[2] = VT7205I2CBus3Init(pScrn->scrnIndex);
    else if (pVia->Chipset > VT7205) {
	pVia->I2CBus[2] = I2CBus3Init(pScrn->scrnIndex);
	pVia->I2CBus[3] = I2CBus4Init(pScrn->scrnIndex);
	pVia->I2CBus[4] = I2CBus5Init(pScrn->scrnIndex);
    }
}

/*
 * Sure, it is polluting namespace, but this one is quite useful.
 */
Bool
xf86I2CMaskByte(I2CDevPtr d, I2CByte subaddr, I2CByte value, I2CByte mask)
{
    I2CByte tmp;
    Bool ret;

    ret = xf86I2CReadByte(d, subaddr, &tmp);
    if (!ret)
	return FALSE;

    tmp &= ~mask;
    tmp |= (value & mask);

    return xf86I2CWriteByte(d, subaddr, tmp);
}
