/*
 * Copyright 2009      Luc Verhaegen.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "xf86xvmc.h"
#include "fourcc.h"

#include "X11/extensions/XvMC.h"

#ifdef BROKEN_VLDXVMC_H
#define XVMC_VLD 0x0020000
#else
#include "X11/extensions/vldXvMC.h"
#endif

#include "via_driver.h"
#include "via_video.h"
#include "via_mpeg.h"
#include "via_xvmc.h"

/*
 *
 * First, our own X protocol to make XvMC less brainless.
 *
 */
#include <extnsionst.h>
#include <dixstruct.h>
#include "lib/xvmc/xvmce_proto.h"

struct XvMCEPrivates
{
    int scrnIndex;
};

/*
 *
 */
static int
XvMCEProcQueryExtension(ClientPtr client)
{
    ExtensionEntry *XvMCE_Ext;
    xvmceQueryExtensionReply rep;

    REQUEST_SIZE_MATCH(xvmceQueryExtensionReq);

    XvMCE_Ext = CheckExtension(XVMCE_EXT_NAME);
    if(!XvMCE_Ext || !XvMCE_Ext->extPrivate) {
        xf86Msg(X_WARNING, "%s: Extension not initialised.\n", __func__);
        return BadMatch;
    }

    rep.type = X_Reply;
    rep.sequenceNumber = client->sequence;
    rep.length =
        (sizeof(xvmceQueryExtensionReply) - sizeof(xGenericReply)) >> 2;
    rep.Version = XVMCE_VERSION;
    rep.Release = XVMCE_RELEASE;

    WriteToClient(client, sizeof(xvmceQueryExtensionReply), (char *)&rep);
    return (client->noClientException);
}

/*
 *
 */
static int
XvMCEProcQMatrixSend(ClientPtr client)
{
    ExtensionEntry *XvMCE_Ext;
    struct XvMCEPrivates *Private;
    ScrnInfoPtr pScrn;
    XvPortPtr Port = NULL;
    int status = Success;

    REQUEST(xvmceQMatrixSendReq);
    REQUEST_AT_LEAST_SIZE(xvmceQMatrixSendReq);

    XvMCE_Ext = CheckExtension(XVMCE_EXT_NAME);
    if (!XvMCE_Ext || !XvMCE_Ext->extPrivate) {
        xf86Msg(X_WARNING, "%s: Extension not initialised.\n", __func__);
        return BadMatch;
    }

    Private = XvMCE_Ext->extPrivate;
    pScrn = xf86Screens[Private->scrnIndex];

    /* Does our client own this port? */
#ifdef NEW_LOOKUPRESOURCE
    status = dixLookupResourceByType((void *) &Port, stuff->Port, XvGetRTPort(),
				     client, DixReadAccess);
#else
    Port = LookupIDByType(stuff->Port, XvGetRTPort());
    if (!Port)
	status = BadAccess;
#endif
    if (status != Success) {
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                   "%s: Client doesn't own this Xv Port.\n", __func__);
        return status;
    }

    /* check size: 16 = 64Bytes / 4, from 8x8 QMatrix */
    if (client->req_len != ((sz_xvmceQMatrixSendReq >> 2) + 16)) {
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                   "%s: Request has wrong length.\n", __func__);
        return BadLength;
    }

    ViaMpegQMatrix(VIAPTR(pScrn)->Swov->Mpeg, stuff->Type, (CARD8 *) &stuff[1]);

    return (client->noClientException);
}

/*
 *
 */
static int
XvMCEProcBufferSliceInit(ClientPtr client)
{
    ExtensionEntry *XvMCE_Ext;
    struct XvMCEPrivates *Private;
    ScrnInfoPtr pScrn;
    XvPortPtr Port = NULL;
    int status = Success;
    CARD8 F_MV_Range, B_MV_Range;
    CARD8 PictureStructure, FrameType, IntraDCPrecision;
    Bool ScanningAlternate, PredictionFrame, TopFieldFirst;
    Bool IntraHasMV, UseQScaleTableType1, UseIntraDCTable1;

    REQUEST(xvmceBufferSliceInitReq);
    REQUEST_SIZE_MATCH(xvmceBufferSliceInitReq);

    XvMCE_Ext = CheckExtension(XVMCE_EXT_NAME);
    if (!XvMCE_Ext || !XvMCE_Ext->extPrivate) {
        xf86Msg(X_WARNING, "%s: Extension not initialised.\n", __func__);
        return BadMatch;
    }

    Private = XvMCE_Ext->extPrivate;
    pScrn = xf86Screens[Private->scrnIndex];

    /* Does our client own this port? */
#ifdef NEW_LOOKUPRESOURCE
    status = dixLookupResourceByType((void *) &Port, stuff->Port, XvGetRTPort(),
				     client, DixReadAccess);
#else
    Port = LookupIDByType(stuff->Port, XvGetRTPort());
    if (!Port)
	status = BadAccess;
#endif
    if (status != Success) {
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                   "%s: Client doesn't own this Xv Port.\n", __func__);
        return BadAccess;
    }

    F_MV_Range = (stuff->MVRange_ForwardsHorizontal & 0x0F) |
	((stuff->MVRange_ForwardsVertical & 0x0F) << 4);
    B_MV_Range = (stuff->MVRange_BackwardsHorizontal & 0x0F) |
	((stuff->MVRange_BackwardsVertical & 0x0F) << 4);

    switch (stuff->PictureType & XVMCE_PICTURETYPE_STRUCTURE_MASK) {
    case XVMCE_PICTURETYPE_FIELD_TOP:
	PictureStructure = MPEG_PICTURE_FIELD_TOP;
	break;
    case XVMCE_PICTURETYPE_FIELD_BOTTOM:
	PictureStructure = MPEG_PICTURE_FIELD_BOTTOM;
	break;
    case XVMCE_PICTURETYPE_FRAME:
	PictureStructure = MPEG_PICTURE_FRAME;
	break;
    default:
	PictureStructure = 0;
	break;
    }

    switch (stuff->PictureType & XVMCE_PICTURETYPE_FRAMETYPE_MASK) {
    case XVMCE_PICTURETYPE_FRAMETYPE_I:
	FrameType = MPEG_FRAMETYPE_I;
	break;
    case XVMCE_PICTURETYPE_FRAMETYPE_P:
	FrameType = MPEG_FRAMETYPE_P;
	break;
    case XVMCE_PICTURETYPE_FRAMETYPE_B:
	FrameType = MPEG_FRAMETYPE_B;
	break;
    default:
	FrameType = 0;
	break;
    }

    if (stuff->PictureType & XVMCE_PICTURETYPE_SCAN_ALTERNATE)
	ScanningAlternate = TRUE;
    else
	ScanningAlternate = FALSE;

    if (stuff->SliceFlags & XVMCE_SLICEFLAGS_PREDICTION_FRAME)
	PredictionFrame = TRUE;
    else
	PredictionFrame = FALSE;

    if (stuff->SliceFlags & XVMCE_SLICEFLAGS_FIELD_FIRST_TOP)
	TopFieldFirst = TRUE;
    else
	TopFieldFirst = FALSE;

    if (stuff->SliceFlags & XVMCE_SLICEFLAGS_INTRA_HAS_MV)
	IntraHasMV = TRUE;
    else
	IntraHasMV = FALSE;

    if (stuff->SliceFlags & XVMCE_SLICEFLAGS_QUANTIZATION_TABLE_MASK)
	UseQScaleTableType1 = TRUE;
    else
	UseQScaleTableType1 = FALSE;

    IntraDCPrecision = (stuff->SliceFlags & XVMCE_SLICEFLAGS_INTRA_PRECISION_MASK) >> 4;

    if (stuff->SliceFlags & XVMCE_SLICEFLAGS_INTRA_DC_TABLE_MASK)
	UseIntraDCTable1 = TRUE;
    else
	UseIntraDCTable1 = FALSE;

    MPEGDebug(ViaDebug(pScrn->scrnIndex, "%s: Buffer 0x%08X\n", __func__,
		       stuff->Buffer));
#if 0
    ViaDebug(pScrn->scrnIndex, "\t PictureType: 0x%02X, SliceFlags: 0x%02X\n",
	     stuff->PictureType, stuff->SliceFlags);
    ViaDebug(pScrn->scrnIndex, "\t Present %01X: backwards 0x%08X, forwards 0x%08X\n",
	     stuff->ReferencesPresent, stuff->BackwardsReferenceBuffer,
	     stuff->ForwardsReferenceBuffer);
    ViaDebug(pScrn->scrnIndex, "\t Ranges BH %d, BV %d, FH %d, FV %d\n",
	     stuff->MVRange_BackwardsHorizontal, stuff->MVRange_BackwardsVertical,
	     stuff->MVRange_ForwardsHorizontal, stuff->MVRange_ForwardsVertical);
#endif

    status = ViaMpegSliceInit(VIAPTR(pScrn)->Swov->Mpeg, stuff->Buffer,
			      stuff->ForwardsReferenceBuffer,
			      stuff->BackwardsReferenceBuffer,
			      F_MV_Range, B_MV_Range, PictureStructure,
			      FrameType, ScanningAlternate, PredictionFrame,
			      TopFieldFirst, IntraHasMV, UseQScaleTableType1,
			      IntraDCPrecision, UseIntraDCTable1);
    if (status != Success)
	return status;

    return (client->noClientException);
}

/*
 *
 */
static int
XvMCEProcBufferSliceSend(ClientPtr client)
{
    ExtensionEntry *XvMCE_Ext;
    struct XvMCEPrivates *Private;
    ScrnInfoPtr pScrn;
    XvPortPtr Port = NULL;
    int status = Success;

    REQUEST(xvmceBufferSliceSendReq);
    REQUEST_AT_LEAST_SIZE(xvmceBufferSliceSendReq);

    XvMCE_Ext = CheckExtension(XVMCE_EXT_NAME);
    if (!XvMCE_Ext || !XvMCE_Ext->extPrivate) {
        xf86Msg(X_WARNING, "%s: Extension not initialised.\n", __func__);
        return BadMatch;
    }

    Private = XvMCE_Ext->extPrivate;
    pScrn = xf86Screens[Private->scrnIndex];

    /* Does our client own this port? */
#ifdef NEW_LOOKUPRESOURCE
    status = dixLookupResourceByType((void *) &Port, stuff->Port, XvGetRTPort(),
				     client, DixReadAccess);
#else
    Port = LookupIDByType(stuff->Port, XvGetRTPort());
    if (!Port)
	status = BadAccess;
#endif
    if (status != Success) {
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                   "%s: Client doesn't own this Xv Port.\n", __func__);
        return status;
    }

    ViaMpegSlicePut(VIAPTR(pScrn)->Swov->Mpeg, stuff->Buffer, stuff->SliceCount,
		    stuff->SliceLength, (CARD8 *) &stuff[1]);

    return (client->noClientException);
}

/*
 *
 */
static int
XvMCEHandler(ClientPtr client)
{
    REQUEST(xReq);

    switch (stuff->data) {
    case xvmce_QueryExtension:
        return XvMCEProcQueryExtension(client);
    case xvmce_QMatrixSend:
	return XvMCEProcQMatrixSend(client);
    case xvmce_BufferSliceInit:
	return XvMCEProcBufferSliceInit(client);
    case xvmce_BufferSliceSend:
	return XvMCEProcBufferSliceSend(client);
    default:
        break;
    }

    return BadRequest;
}

/*
 *
 */
static int
XvMCESHandler(ClientPtr client)
{
    ErrorF("%s is not implemented.\n", __func__);

    return BadRequest;
}

/*
 *
 */
static void
XvMCEClose(ExtensionEntry *XvMCE_Ext)
{
    if (XvMCE_Ext->extPrivate) {
        free(XvMCE_Ext->extPrivate);
        XvMCE_Ext->extPrivate = NULL;
    }
}

/*
 *
 */
static Bool
XvMCEInit(ScrnInfoPtr pScrn)
{
    ExtensionEntry *XvMCE_Ext;
    struct XvMCEPrivates *Private;

    VIAFUNC(pScrn);

    XvMCE_Ext = AddExtension(XVMCE_EXT_NAME, 0, 0, XvMCEHandler, XvMCESHandler,
			     XvMCEClose, StandardMinorOpcode);
    if (!XvMCE_Ext) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "%s: Failed to add %s.\n",
                   __func__, XVMCE_EXT_NAME);
        return FALSE;
    }

    Private = xnfcalloc(sizeof(struct XvMCEPrivates), 1);
    Private->scrnIndex = pScrn->scrnIndex;
    XvMCE_Ext->extPrivate = Private;

    return TRUE;
}

/*
 *
 * Now, use what is somewhat useable from XvMC.
 *
 */

/*
 *
 */
static int
ViaMCCreateContext(ScrnInfoPtr pScrn, XvMCContextPtr context,
		     int *num_priv, CARD32 **priv)
{
    struct ViaMpeg *Mpeg = VIAPTR(pScrn)->Swov->Mpeg;

    ViaDebug(pScrn->scrnIndex, "%s: adapter %d, surfaceid %08X (%dx%d); %08X\n",
	     __func__, context->adapt_num, context->surface_type_id,
	     context->width, context->height, context->flags);

    if (Mpeg->ContextID)
	return BadAccess;

    if (!ViaMpegEngineInit(Mpeg, context->width, context->height))
	return BadValue;

    Mpeg->ContextID = context->context_id;

    return Success;
}

/*
 *
 */
static void
ViaMCDestroyContext(ScrnInfoPtr pScrn, XvMCContextPtr context)
{
    struct ViaMpeg *Mpeg = VIAPTR(pScrn)->Swov->Mpeg;

    VIAFUNC(pScrn);

    if (Mpeg->ContextID != context->context_id)
	return;

    Mpeg->ContextID = 0;
}

/*
 *
 */
static int
ViaMCBufferCreate(ScrnInfoPtr pScrn, XvMCSurfacePtr Buffer,
		  int *num_priv, CARD32 **priv)
{
    struct ViaMpeg *Mpeg = VIAPTR(pScrn)->Swov->Mpeg;

    VIAFUNC(pScrn);

    if (Mpeg->ContextID != Buffer->context->context_id)
	return BadAccess;

    if (!ViaMpegBufferGrab(Mpeg, Buffer->surface_id))
	return BadAlloc;

    return Success;
}

/*
 *
 */
static void
ViaMCBufferDestroy(ScrnInfoPtr pScrn, XvMCSurfacePtr Buffer)
{
    struct ViaMpeg *Mpeg = VIAPTR(pScrn)->Swov->Mpeg;

    VIAFUNC(pScrn);

    if (Mpeg->ContextID != Buffer->context->context_id)
	return;

    ViaMpegBufferRelease(Mpeg, Buffer->surface_id);
}

/*
 * This whole XvMC-VLD crap is soo broken. Not only do we implement subpictures
 * horribly, we also have a dependance on them. So fool the implementations
 * that we do have subpictures, but fail to allocate them later on.
 */
static int ViaMCImageIDsBrokenXine[1] = {FOURCC_AI44};
static XF86MCImageIDList ViaMCImageIDListBrokenXine = {1, ViaMCImageIDsBrokenXine};

static XF86ImageRec ViaMCAI44BrokenXine = XVIMAGE_AI44;
static XF86ImagePtr ViaMCSubPicturesBrokenXine[1] = {&ViaMCAI44BrokenXine};

static XF86MCSurfaceInfoRec ViaMCMpeg2SurfaceBrokenXine = {
    FOURCC_YV12,
    XVMC_CHROMA_FORMAT_420,
    0,
    1024, 1024, 1024, 1024,
    XVMC_MPEG_2 | XVMC_VLD,
    XVMC_OVERLAID_SURFACE,
    &ViaMCImageIDListBrokenXine
};

static XF86MCSurfaceInfoPtr ViaMCSurfacesBrokenXine[1] = {&ViaMCMpeg2SurfaceBrokenXine};

static XF86MCAdaptorRec ViaMCAdaptorBrokenXine = {
	ViaXVAdaptorName,
	1, ViaMCSurfacesBrokenXine,
	1, ViaMCSubPicturesBrokenXine,
	ViaMCCreateContext,
	ViaMCDestroyContext,
	ViaMCBufferCreate,
	ViaMCBufferDestroy,
	NULL,
	NULL
};
static XF86MCAdaptorPtr ViaMCAdaptorsBrokenXine[1] = {&ViaMCAdaptorBrokenXine};

/*
 * Working XvMC clients that do not need crappily implemented subpictures.
 */
static XF86MCSurfaceInfoRec ViaMCMpeg2Surface = {
    FOURCC_YV12,
    XVMC_CHROMA_FORMAT_420,
    0,
    1024, 1024, 1024, 1024,
    XVMC_MPEG_2 | XVMC_VLD,
    XVMC_OVERLAID_SURFACE,
    NULL
};
static XF86MCSurfaceInfoPtr ViaMCSurfaces[1] = {&ViaMCMpeg2Surface};

static XF86MCAdaptorRec ViaMCAdaptor = {
	ViaXVAdaptorName,
	1, ViaMCSurfaces,
	0, NULL,
	ViaMCCreateContext,
	ViaMCDestroyContext,
	ViaMCBufferCreate,
	ViaMCBufferDestroy,
	NULL,
	NULL
};
static XF86MCAdaptorPtr ViaMCAdaptors[1] = {&ViaMCAdaptor};

/*
 * Implement an option to allow broken Xine clients to be fooled.
 */
static Bool
ViaMCBrokenXine(int scrnIndex)
{
    enum ViaMCOpts {OPTION_BROKENXINE};
    static OptionInfoRec ViaMCOptions[] =
	{
	    {OPTION_BROKENXINE,  "XvMCBrokenXine",  OPTV_BOOLEAN, {0}, FALSE},
	    {-1,                 NULL,          OPTV_NONE,    {0}, FALSE}
	};
    OptionInfoPtr  Options;
    Bool ret;

    Options = xnfalloc(sizeof(ViaMCOptions));
    memcpy(Options, ViaMCOptions, sizeof(ViaMCOptions));

    xf86ProcessOptions(scrnIndex, xf86Screens[scrnIndex]->options, Options);

    ret = xf86ReturnOptValBool(Options, OPTION_BROKENXINE, FALSE);

    free(Options);
    return ret;
}

/*
 *
 */
Bool
ViaXvMCInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
{
    int ret;

    VIAFUNC(pScrn);

    if (!XvMCEInit(pScrn)) {
	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
		   "%s: Failed to initialise XvMCE extension.\n", __func__);
	return FALSE;
    }

    if (ViaMCBrokenXine(pScrn->scrnIndex)) {
	xf86DrvMsg(pScrn->scrnIndex, X_INFO,
		   "%s: working around broken XINE XXMC module.\n", __func__);
	ret = xf86XvMCScreenInit(pScreen, 1, ViaMCAdaptorsBrokenXine);
    } else
	ret = xf86XvMCScreenInit(pScreen, 1, ViaMCAdaptors);

    if (!ret) {
	xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
		   "%s: Failed to initialise XvMC.\n", __func__);
	return FALSE;
    }

    xf86XvMCRegisterDRInfo(pScreen, "XvMCunichrome", "1:0.0", 0, 0, 0);

    return TRUE;
}

/*
 *
 */
void
ViaXvMCDestroy(ScrnInfoPtr pScrn)
{
    VIAFUNC(pScrn);
}
